import os
import sys
from pathlib import Path
from urllib.request import urlopen

import requests

import plateNet.color as color
from plateNet.tools import message


def internet_on():
    try:
        urlopen('https://www.google.com/', timeout=1)
        return True
    except:
        return False


def __download_file_from_google_drive(id, destination):
    if internet_on():
        URL = "https://docs.google.com/uc?export=download"

        session = requests.Session()

        response = session.get(URL, params={'id': id}, stream=True)
        token = __get_confirm_token(response)

        if token:
            params = {'id': id, 'confirm': token}
            response = session.get(URL, params=params, stream=True)

        __save_response_content(response, destination)
        return True
    else:
        message("Could not download because no internet connection was found", color.CRED)
        sys.exit()


def __get_confirm_token(response):
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            return value

    return None


def __save_response_content(response, destination):
    CHUNK_SIZE = 327680

    with open(destination, "wb") as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk:
                f.write(chunk)

    return True


def download_plateNet_448(update:bool=False,savefolder="plateNet-448"):
    # savefolder = "plateNet-448"

    if not os.path.exists(savefolder):
        os.mkdir(savefolder)

    weights = "plateNet_448_last.weights"
    # weightsID = '1ZFCA_LtB3F7hk71a9YWMRSFeGZGgfiBr'
    weightsID = '1tjfTEZYcfe3dw91mOL7yNa-EejTQxRfd'
    # https://drive.google.com/open?id=1tjfTEZYcfe3dw91mOL7yNa-EejTQxRfd
    cfg = "plateNet_448.cfg"
    cfgID = '1V7etbj5YB41AcPH4s4hZQ_YVVIBnOZvC'
    # https://drive.google.com/open?id=14UBqeNs-yxboHeYhS-98qJMFsOFlEK6Z
    isr = 'rdn-C6-D20-G64-G064-x2_ArtefactCancelling_epoch219.hdf5'
    isrID='14UBqeNs-yxboHeYhS-98qJMFsOFlEK6Z'
    items =[
        [cfg, cfgID],
        [weights,weightsID],
        [isr,isrID]
    ]
    downloadedPaths = []
    for item in items:

        path = os.path.join(savefolder, item[0])
        downloadedPaths.append(path)
        if not Path(path).is_file() or update:
            message("%s downloading..." % path, color.CVIOLET)
            __download_file_from_google_drive(item[1], path)
        else:
            message("%s found." % path, color.CGREEN)

    return downloadedPaths



import os
from pathlib import Path

import cv2
import imutils

import plateNet.color as color
from plateNet.folder import generateFolder
from plateNet.tools import message


class DatasetManager:
    def __init__(self, imgList: list):
        self.imgList = imgList

    def create(self, width: int=448,
               angle: float = .0,
               savefolder:str=None):
        message("preparing dataset...", color.CYELLOW2)

        resultList = []
        for imgPath in self.imgList:
            # print("img : "+imgPath)
            img = cv2.imread(imgPath)
            rotatedImg = imutils.rotate(img, angle)
            resizedImg = imutils.resize(rotatedImg, width)
            resultList.append(resizedImg)
            if savefolder is not None:
                # mainPath = os.path.join(imgPath.split("/")[:-1])

                mainPath = generateFolder(Path(imgPath).parent, savefolder)
                filename = imgPath.split("/")[-1]
                savename = os.path.join(mainPath, filename)
                cv2.imwrite(savename, resizedImg)
        message("dataset ready to use...", color.CYELLOW2)
        return resultList



from distutils.core import setup

setup(
    name='plateNet',  # How you named your package folder (MyLib)
    packages=['plateNet'],  # Chose the same as "name"
    version='0.4.0',  # Start with a small number and increase it with every change you make
    license='MIT',  # Chose a license from here:
    description='plate detect with Yolov3',  # Give a short description about your library
    author='Hacı',  # Type in your name
    author_email='karakus.haciveli@gmail.com',  # Type in your E-Mail
    install_requires=[  # I get to this in a second
        'opencv-python',
        'imutils',
        'numpy',
        'ISR',
        'requests'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',  # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',  # Again, pick a license
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



import json
import urllib.request
import xml.etree.cElementTree as ET

import cv2
import numpy as np
import requests
from ISR.models import RDN
from PIL import Image
from plateNet.folder import generateOutputFolder
from plateNet.tools import getTime


class Saver:

    def __init__(self, imagePath, savefolder="plates"):
        self.imagePath = imagePath
        self.savefolder = savefolder
        self.mainPath, self.filename = generateOutputFolder(imagePath, savefolder)

    def saveJSON(self, boxes, confidence):
        json_fileName = "{}/{}.json".format(self.mainPath, self.filename)
        boxes = {
            'Date': getTime(),
            'File': "{}".format(self.imagePath),
            'Confidence': confidence,
            'X': boxes[0],
            'Y': boxes[1],
            'W': boxes[2],
            'H': boxes[3]
        }
        with open(json_fileName, 'w') as json_file:
            json.dump(boxes, json_file, ensure_ascii=False, indent=4)

    def saveXML(self, boxes, confidence):
        xml_fileName = "{}/{}.xml".format(self.mainPath, self.filename)

        root = ET.Element("PlateNet")
        features = ET.SubElement(root, "features")

        ET.SubElement(features, "date", name="detect time").text = getTime()
        ET.SubElement(features, "file", name="input image").text = self.imagePath
        ET.SubElement(features, "conf", name="confidence").text = str(confidence)

        position = ET.SubElement(features, "position")
        ET.SubElement(position, "x", name="x coord").text = str(boxes[0])
        ET.SubElement(position, "y", name="y coord").text = str(boxes[1])
        ET.SubElement(position, "w", name="width").text = str(boxes[2])
        ET.SubElement(position, "h", name=" height").text = str(boxes[3])

        tree = ET.ElementTree(root)
        tree.write(xml_fileName)

    def savePNG(self, img):
        if img is not None:
            imgName = "{}/{}.png".format(self.mainPath, self.filename)
            self.__ISR(img, imgName)

    def __SRGAN(self, img, name):
        headers = {
            'api-key': '851d198d-097b-4885-bc5b-20c0b86f689f',
        }

        cv2.imwrite(name, img)

        files = {
            'image': open(name, "rb"),
        }

        response = requests.post('https://api.deepai.org/api/torch-srgan', headers=headers, files=files)
        url = response.json()["output_url"]

        urllib.request.urlretrieve(url, name)

    def __ISR(self, img, name):
        height,width = img.shape[:2]
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = Image.fromarray(img)
        lr_img = np.array(img)

        rdn = RDN(arch_params={'C': 6, 'D': 20, 'G': 64, 'G0': 64, 'x': 2})
        rdn.model.load_weights('plateNet-448/rdn-C6-D20-G64-G064-x2_ArtefactCancelling_epoch219.hdf5')
        # rdn.model.load_weights('app/plateNet-448/rdn-C6-D20-G64-G064-x2_ArtefactCancelling_epoch219.hdf5')

        sr_img = rdn.predict(lr_img)
        res = Image.fromarray(sr_img)
        new_height = 100
        new_width = int(new_height * width/ height)
        res = res.resize((new_width,new_height),Image.ANTIALIAS)
        res.save(name)

    # def saver(currentImg,plate,boxes):

    # savePNG(main_path,currentImg,plate)
    # saveJSON(main_path,currentImg,boxes[0])



import os

from plateNet.tools import message

image_types = (".jpg", ".jpeg", ".png", ".bmp", ".tif", ".tiff")


def imageFolder2List(basePath, ignoreFolders=None):
    # return the set of files that are valid
    return list(__list_files__(basePath, ignoreFolders=ignoreFolders))


def __list_files__(basePath, ignoreFolders):
    for (rootDir, dirNames, filenames) in os.walk(basePath):

        for filename in filenames:
            if ignoreFolders is not None and rootDir.split("/")[-1] in ignoreFolders:
                continue
            ext = filename[filename.rfind("."):].lower()

            if image_types is None or ext.endswith(image_types):
                imagePath = os.path.join(rootDir, filename)
                yield imagePath


def generateFolder(path, foldername):
    mainPath = os.path.join(path, foldername)

    if not os.path.exists(mainPath):
        os.mkdir(mainPath)
        message("{} folder generated.".format(mainPath), color.CGREEN)

    return mainPath


def generateOutputFolder(path, foldername):
    head, root = os.path.split(path)

    filename = root.split('.')[0]
    mainPath = "{}/{}".format(head, foldername)

    if not os.path.exists(mainPath):
        os.mkdir(mainPath)

    return mainPath, filename



import sys
from operator import itemgetter

import cv2
import imutils
import numpy as np
import plateNet.color as color
from plateNet.save import Saver
from plateNet.tools import extractPlate
from plateNet.tools import message


class PlateNetDetector:
    def __init__(self, paths: list):

        message("reading model..", color.CGREEN)
        try:
            self.net = cv2.dnn.readNetFromDarknet(paths[0], paths[1])
            self.layer_names = [self.net.getLayerNames()[index[0] - 1] for index in self.net.getUnconnectedOutLayers()]
        except:
            message("failed to load model", color.CRED)
            sys.exit()

    def fromList(self, imgList: list, angleRange: tuple = (-5, 15), size: int = 448, unitAngle: float = .1,
                 expected_confidence: float = .5, save=True):

        startAngle, stopAngle = angleRange
        startAngle = startAngle * 10
        stopAngle = stopAngle * 10
        message("detecting plates...", color.CYELLOW2)
        if len(imgList) > 0:
            message("{} images found".format(len(imgList)), color.CBOLD)
            for currentImg in imgList:
                origImg = cv2.imread(currentImg)
                combinations_img = []
                for angle in range(startAngle, stopAngle):
                    rotated_img = imutils.rotate(origImg, angle * unitAngle)
                    detected_prob = self.__detect(rotated_img, expected_confidence, size)
                    combinations_img.append(detected_prob)

                plate, confidences, boxes = max(combinations_img, key=itemgetter(1))
                message("[{}/{}] {} : {}".format(imgList.index(currentImg) + 1,
                                                 len(imgList),
                                                 currentImg,
                                                 confidences), color.CBOLD)

                if len(boxes) != 0:
                    if save:
                        saver = Saver(currentImg)
                        saver.savePNG(plate)
                        saver.saveJSON(boxes[0], confidences[0])
                        saver.saveXML(boxes[0], confidences[0])
                    else:
                        return plate, confidences, boxes

    def __detect(self, image, expected_confidence, IMG_SIZE):
        boxes = []
        confidences = []
        # H = Height of image
        # W = width of image
        (H, W) = image.shape[:2]
        blob = cv2.dnn.blobFromImage(image, 1 / 255.0, (IMG_SIZE, IMG_SIZE), swapRB=True, crop=False)
        self.net.setInput(blob)
        layer_outputs = self.net.forward(self.layer_names)

        for output in layer_outputs:
            for detection in output:
                scores = detection[5:]
                classID = np.argmax(scores)
                confidence = scores[classID]
                # take care when higher ... else skip this
                if confidence > expected_confidence:
                    # print("founded plate confidences : %.2f" % confidence)
                    box = detection[0:4] * np.array([W, H, W, H])
                    (centerX, centerY, width, height) = box.astype("int")
                    # tespit edilen plakanın konumları
                    x = int(centerX - (width / 2))
                    y = int(centerY - (height / 2))

                    boxes.append([x, y, int(width), int(height)])
                    confidences.append(float(confidence))

        idxs = cv2.dnn.NMSBoxes(boxes, confidences, score_threshold=0.5, nms_threshold=0.5)

        plateArea = extractPlate(image, boxes, idxs)

        return plateArea, confidences, boxes



import sys
from datetime import datetime

import cv2
import numpy as np
import imutils
import plateNet.color as color


def merge(first, second):
    """
    merge two images horizontally
    :param first:image(numpy array) will on the left
    :param second:image(numpy array) will on the right
    :return:
    """
    ha, wa = first.shape[:2]
    hb, wb = second.shape[:2]
    max_height = np.max([ha, hb])
    total_width = wa + wb
    new_img = np.zeros(shape=(max_height, total_width, 3))
    new_img[:ha, :wa] = first
    new_img[:hb, wa:wa + wb] = second

    return new_img


def contrast(image, clipLimit=4.0, tileGridSize=(8, 8)):
    """
    Apply more contrast to image
    :param image: image(numpy array)
    :param clipLimit: Threshold for contrast limiting.
    :param tileGridSize: Size of grid for histogram equalization. Input image will be divided into
                         equally sized rectangular tiles. tileGridSize defines the number of tiles in row and column
    :return:
    """
    lab = cv2.cvtColor(image, cv2.COLOR_BGR2LAB)
    l, a, b = cv2.split(lab)

    clahe = cv2.createCLAHE(clipLimit=clipLimit, tileGridSize=tileGridSize)
    cl = clahe.apply(l)

    limg = cv2.merge((cl, a, b))
    final = cv2.cvtColor(limg, cv2.COLOR_LAB2BGR)

    return final


def getPlate(origImg, currentImg, coord):
    # coord[0] = x
    # coord[1] = y
    # coord[2] = w
    # coord[3] = h
    x_rate = origImg.shape[0] / currentImg.shape[0]
    y_rate = origImg.shape[1] / currentImg.shape[1]

    new_x = int(coord[0] * x_rate)
    new_y = int(coord[1] * y_rate)
    new_w = int(coord[2] * x_rate)
    new_h = int(coord[3] * y_rate)

    return origImg[new_y - 1:new_y + new_h + 1, new_x - 1:new_x + new_w + 1]

def moph(img):
    hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
    hue, saturation, value = cv2.split(hsv)
    # cv2.imshow('gray', value)
    # cv2.imwrite(temp_folder + '2 - gray.png', value)

    # kernel to use for morphological operations
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (3, 3))

    # applying topHat/blackHat operations
    topHat = cv2.morphologyEx(value, cv2.MORPH_TOPHAT, kernel)
    blackHat = cv2.morphologyEx(value, cv2.MORPH_BLACKHAT, kernel)
    # cv2.imshow('topHat', topHat)
    # cv2.imshow('blackHat', blackHat)
    # cv2.imwrite(temp_folder + '3 - topHat.png', topHat)
    # cv2.imwrite(temp_folder + '4 - blackHat.png', blackHat)

    # add and subtract between morphological operations
    add = cv2.add(value, topHat)
    subtract = cv2.subtract(add, blackHat)
    return cv2.cvtColor(subtract,cv2.COLOR_GRAY2BGR)

def extractPlate(img, boxes, idxs):
    if len(idxs) > 0:
        for i in idxs.flatten():
            (x, y) = (boxes[i][0], boxes[i][1])
            (w, h) = (boxes[i][2], boxes[i][3])

            res= img[y:y + h, x:x + w]
            # return img[y:y + h, x:x + w]
            # return img[y-50:y + h+50, x-50:x + w+50]
            return res

def plot(orig_img, rotated_img, mboxes, confidences, idxs, plate_ratio=1, expected_confidence=.5, thickness=1):
    (H, W) = rotated_img.shape[:2]

    plate_area = None

    if len(idxs) > 0:
        for i in idxs.flatten():
            # tespit edilen plaka özellikleri
            (x, y) = (mboxes[i][0], mboxes[i][1])
            (w, h) = (mboxes[i][2], mboxes[i][3])

            # image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
            # apply_lab(image)
            # image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)

            if confidences[i] > expected_confidence:
                # sol üst
                cv2.line(rotated_img,
                         pt1=(x, y),
                         pt2=(0, H - plate_ratio * h),
                         color=color.orange,
                         thickness=thickness)
                # sağ üst
                cv2.line(rotated_img,
                         pt1=(x + w, y),
                         pt2=(plate_ratio * w, H - plate_ratio * h),
                         color=color.orange,
                         thickness=thickness)
                # sağ alt
                cv2.line(rotated_img,
                         pt1=(x + w, y + h),
                         pt2=(plate_ratio * w, H),
                         color=color.orange,
                         thickness=thickness)
                # sol alt
                cv2.line(rotated_img,
                         pt1=(x, y + h),
                         pt2=(0, H),
                         color=color.orange,
                         thickness=thickness)
                # original halinden bulunan plaka alanını al
                plate_area = orig_img[
                             y:y + h,
                             x:x + w]
                # # find_contours(plate_area)
                # plate_area = getPlate(rotated_img, [H, W], [x, y, w, h])
                # plakayı plate_ratio oranında büyülterek sol altta göster
                rotated_img[H - plate_ratio * h:H, 0:plate_ratio * w] = cv2.resize(plate_area, (plate_ratio * w,
                                                                                                plate_ratio * h))
                # Büyük formdaki plakaya çerçeve yap
                cv2.rectangle(rotated_img,
                              pt1=(0, H - plate_ratio * h),
                              pt2=(plate_ratio * w, H),
                              color=color.orange,
                              thickness=thickness)
                # confidence değeri arkaplanı
                cv2.rectangle(rotated_img, (x, y - h), (x + w, y), color.black, cv2.FILLED)
                # confidence değeri
                text = "%{:.2f}".format(confidences[i] * 100)
                cv2.putText(rotated_img, text, (x, y - 5), cv2.FONT_HERSHEY_SIMPLEX, 0.4, (0, 150, 250), 1)

                # plaka çerçevesi
                cv2.rectangle(rotated_img, (x, y), (x + w, y + h), color.orange, thickness)

    else:
        cv2.line(rotated_img, (0, 0), (rotated_img.shape[1], rotated_img.shape[0]), color.red, 5)
        cv2.line(rotated_img, (rotated_img.shape[1], 0), (0, rotated_img.shape[0]), color.red, 5)

    return rotated_img, plate_area


def getTime():
    """
    get current time formatted type
    :return:
    """
    now = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

    return now


def message(message: str, color=color.CBOLD):
    """
    print colored text message
    :param message: message to be written
    :param color:message text color from
    :return:
    """
    sys.stderr.write("\n%s" % color + "[%s] " % getTime() + message.strip() + '\x1b[0m\n')
